//---------------------------------------------------------------------------
#include <vcl.h>
#include <NukeDX.h>
#pragma hdrstop

//---------------------------------------------------------------------------
#pragma package(smart_init)

#include "Unit1.h"
#include "Objects.h"
#include "Game.h"

// ***    ***
extern NDX_Screen  *NdxScreen;
extern NDX_Surface *gameArt;
extern NDX_Input   *NdxInput;
extern Game_Control *GameControl;

extern NDX_Sample *sampleWalk;
extern NDX_Sample *sampleMy;
extern NDX_Sample *sampleDie;
extern NDX_Sample *sampleGameOver;
extern NDX_Sample *sampleFall;

extern BOOL bTesting;
extern int Program_State;

extern double PlayerDelay;

extern Game_Control *GameControl;
NDX_SpriteControl *Sprites;

GO_Level *level;    //    
GO_Player *player;  //   


/*          */

void OutMessage(String s)
{
    //   
    static int n=0;
    Form1->MessageMemo->Lines->Add(IntToStr(n)+"  "+s);
    n++;
}

BOOL HitRect(const RECT *Rect1,const RECT *Rect2)
{
	if(Rect1->top >= Rect2->bottom) return FALSE;
	else if(Rect1->bottom <= Rect2->top) return FALSE;
	else if(Rect1->right <= Rect2->left) return FALSE;
	else if(Rect1->left >= Rect2->right) return FALSE;

	return TRUE;
}




//  ***    ***

//  
Game_Object::Game_Object()
{
    //  
    x=y=0;
    frame=0;
    div_x=div_y=0;

    len=hgt=16;  //   
    action=SA_STILL;
}

//   
void Game_Object::Create(LEVFILEOBJECT *levobj)
{
    //      
    x=levobj->posx;
    y=levobj->posy;
    param=levobj->param;  //      
}


void Game_Object::Draw()
{
    // -1  ,     (  )
    if( frame != -1 )
    {
        //   frame  framelist'
        Frames.Draw(NdxScreen->Back,x ,y, frame );
    }
}

//  true   
BOOL Game_Object::HitObject( Game_Object* obj1)
{
    Game_Object *obj2=this;

	RECT Rect1 = { obj1->x, obj1->y,
								 obj1->x+obj1->len,
								 obj1->y+obj1->hgt };
	RECT Rect2 = { obj2->x, obj2->y,
								 obj2->x+obj2->len,
								 obj2->y+obj2->hgt };

	return HitRect(&Rect1, &Rect2);
}

//  true      
BOOL Game_Object::is_IntoGround(int _x, int _y)
{
    //     
    if( !level->isPassingTile(_x,_y) ||
        !level->isPassingTile(_x+len,_y) ||
        !level->isPassingTile(_x,_y+hgt-1) ||
        !level->isPassingTile(_x+len,_y+hgt-1) )
            return true;

    return false;
}

BOOL Game_Object::is_IntoGround()
{
    return is_IntoGround(x,y);
}

//  true       
BOOL Game_Object::is_NoGround()
{
    if( !level->isGroundTile(x,y+hgt) &&
        !level->isGroundTile(x+len, y+hgt) )
        return true;

    return false;
}

/*==========================================================
        ***   LEVEL    ****/

//   Level,      GameObject
GO_Level::GO_Level(char *name)
{
    if( Level.Load(name) != true )
        MessageBox( Application->Handle, "Erorr load level.", "Level", MB_OK );
        type=GO_LEVEL;
}

void GO_Level::Draw()
{
    static RECT area=Rect(0,0,320,240);
    //    
    Level.Draw(NdxScreen->Back,x,y,area);
}

// ,      
// (    )
int GO_Level::isPassingTile(int pixel_x,int pixel_y)
{
    int i=Level.GetTile(pixel_x,pixel_y);

    if( i==1 || i==(-1) ) return false;  //     - 
    return true;  //  - 
}

//  true,      
int GO_Level::isGroundTile(int pixel_x,int pixel_y)
{
    int i=Level.GetTile(pixel_x,pixel_y);

    if( i==1 || i==2 || i==(-1)) return true;  //    - 
    return false;
}

/*==========================================================

        ***   PLAYER    ***
*/
//   
GO_Player::GO_Player()
{
    bCanClimbUp=false;
    bCanClimbDown=false;
    bLiftUse = false;
    lift = NULL;
    len=11; hgt=15;
    Frames.AddFrames( gameArt, 0,48,  16,16,  3,2 );
    snd_Walk_h=snd_Fall_h=NULL;
    fd=0;
    action = SA_STILL;
}


//    Move    .
//     -   
//  ,   ..    Message
//     .
void GO_Player::Move(double td)
{
    //   ,
    //       
    if( PlayerDelay ) return;

    /*     */
    div_x=0;

    //        
    if( NdxInput->Keys[DIK_LEFT] && !NdxInput->Keys[DIK_RIGHT])
    {
        //     
        // (    )
        div_x=(action==SA_FALL)? -10:-70;

        //     ,  
        if( is_IntoGround( x+div_x*td, y ))
            div_x=0;
        else
            if( action != SA_FALL )
                action=SA_WALK;
    }

    //        
    if( NdxInput->Keys[DIK_RIGHT] && !NdxInput->Keys[DIK_LEFT])
    {
        //     
        // (    )
        div_x=(action==SA_FALL)? 10: 70;

        //     ,  
        if( is_IntoGround( x+div_x*td, y ))
            div_x=0;
        else
            if( action != SA_FALL )
                action=SA_WALK;
    }


    /*     */

    if( action != SA_FALL )
        div_y = 0;

    //     ,     
    if( action != SA_FALL && is_NoGround() && !bLiftUse )
    {
        Game_Message msg={GM_NOGROUND,0,this};
        Message(&msg);
    }

    //          
    if( action == SA_FALL && ( is_IntoGround() || !is_NoGround()) )
    {
        //     ,    
        if( !is_IntoGround() )
        {
            Game_Message msg={GM_GROUND,0,this};
            Message(&msg);
        }
        else
        {
            // ,    ,  
            int temp_y = y;
            while( is_IntoGround( x,temp_y )) temp_y--;
            y = temp_y;
        }
    }


    //    
    if( level->Level.GetTile(x+2,y+hgt-1) ==2 &&
        level->Level.GetTile(x+len-2,y+hgt-1) ==2 )
    {
        //     ,  
        bCanClimbUp=true;
    }
    else
        bCanClimbUp=false;

    //     
    if( level->Level.GetTile(x+2,y+hgt) ==2  &&
        level->Level.GetTile(x+len-2,y+hgt) ==2  )
    {
        //    (   )  ,  
        bCanClimbDown=true;
    }
    else
        bCanClimbDown=false;

    //        
    if( NdxInput->Keys[DIK_UP] && bCanClimbUp )
    {
        //     , 
        if( action != SA_CLIMB )
        {
                //   __     
                int tmp = (int)x%level->Level.GetTileLen();
                x += (tmp > level->Level.GetTileLen()/2)?
                    level->Level.GetTileLen()-tmp : -tmp ;

                //   
                if( snd_Walk_h ) //    NULL
                if( sampleWalk->IsPlaying(snd_Walk_h) )
                    sampleWalk->Stop(snd_Walk_h);
                //   
                if( snd_Fall_h ) //    NULL
                if( sampleFall->IsPlaying(snd_Fall_h) )
                    sampleFall->Stop(snd_Fall_h);
        }

        action=SA_CLIMB; //      ,   
        div_y=-40;
    }

    //       ,   
    if( NdxInput->Keys[DIK_DOWN] && bCanClimbDown)
    {
        //     , 
        if(action!=SA_CLIMB)
        {
                //   __     
                int tmp = (int)x%level->Level.GetTileLen();
                x += (tmp > level->Level.GetTileLen()/2)?
                    level->Level.GetTileLen()-tmp : -tmp ;

                //   
                if( snd_Walk_h ) //    NULL
                if( sampleWalk->IsPlaying(snd_Walk_h) )
                    sampleWalk->Stop(snd_Walk_h);
                //   
                if( snd_Fall_h ) //    NULL
                if( sampleFall->IsPlaying(snd_Fall_h) )
                    sampleFall->Stop(snd_Fall_h);
        }

        action=SA_CLIMB; //      ,   
        div_y=40;
    }

    /*      */

    //    ,  
    if( action == SA_WALK && div_x==0 )
        action=SA_STILL;

    //    ,  
    if( action == SA_CLIMB && div_y==0 )
        action=SA_STILL;

/*         
     2 .     frame1  frame2.
     ,      , 
        , 

    NDX_FrameList *cur_animation;

       ,    frame1  frame2
         ,  :

        case SA_FALL:
            cur_animation = &animation_player_fall;
            break;
*/
    //    
    switch( action )
    {
        case SA_STILL:
            frame1=0;
            frame2=0;
            break;
        case SA_WALK:
            if( div_x < 0 )
            {
                frame1=4;
                frame2=5;
            }
            else
            {
                frame1=1;
                frame2=2;
            }
            break;
        case SA_FALL:
            frame1=0;
            frame2=3;
            break;
        case SA_CLIMB:
            frame1=1;
            frame2=3;
            break;

    }

    //   
    if( frame == frame1 ) {
        fd+=td;
        if( fd > 0.04 ) {
            frame=frame2;
            fd=0;
        }
    } else
    if( frame == frame2 ) {
        fd+=td;
        if( fd > 0.04 ) {
            frame=frame1;
            fd=0;
        }
    } else {
        frame=frame1;
        fd=0;
    }

    //  
    if( action==SA_FALL )
        frame=3;

    /*     */
    prev_x = x;
    prev_y = y;

    x = x + div_x*td;
    y = y + div_y*td;

    if( is_IntoGround() )
    {
        x = prev_x;
        y = prev_y;
        frame=0;
        action=SA_STILL;
OutMessage("Player Message: GM_INTOGROUND");
    }
}

void GO_Player::Message(void *m)
{
    if( PlayerDelay ) return;

    Game_Message *msg=(Game_Message *)m;
    switch (msg->type)
    {
      case GM_NOGROUND:  //  
OutMessage("Player Message: GM_NOGROUND");
            action=SA_FALL;
            div_x=0;
            div_y=70;
            //   
            if( snd_Walk_h ) //    NULL
            if( sampleWalk->IsPlaying(snd_Walk_h) )
                sampleWalk->Stop(snd_Walk_h);
            //   
            sampleFall->Play(0,&snd_Fall_h);
      break;

      case GM_GROUND:  //  
OutMessage("Player Message: GM_GROUND");
        action=SA_STILL;
        div_x=0;
        div_y=0;
        frame=0;
        //   
        if( snd_Fall_h ) //    NULL
        if( sampleFall->IsPlaying(snd_Fall_h) )
            sampleFall->Stop(snd_Fall_h);
      break;

      // ,    , 
      case GM_LIFTMOVE:
        //     
        if( is_IntoGround(x+msg->sender->div_x, y+msg->sender->div_y) )
        {
OutMessage("Player can't move with lift");
            //     
            //   ,   
            //  ,     
            Game_Message _msg={GM_CANTMOVE,0,this};
            msg->sender->Message(&_msg);
        }
        else
        {
            //   ,   
            //   
            x += lift->div_x;
            y += lift->div_y;
        }
        break;

    }
}

void GO_Player::CheckForHits()
{
/*        ,
          
        ,  ,
      ,   
      GO_Player *player    
          CheckForHits
     .
*/

//   ,  , 
//       
#if 0
    Game_Object *o;
	for(int n=0;n<Sprites->NumSprites;n++)
	{
        o=(Game_Object*)Sprites->Sprites[n];
        switch( o->type )
        {
            case GO_LEVEL:
                break;
            case GO_LIFT:
                break;
            case GO_KEY:
                break;
            case GO_ENEMY1:
                break;
            default: ;
        }
	}
#endif
}

/*==========================================================
        ***   ENEMY1    ****/

GO_Enemy1::GO_Enemy1()
{
    move_count=0;
    len=12; hgt=14;
    Frames.AddFrames( gameArt,0,81,len,hgt,4,1);
}

void GO_Enemy1::Move(double td)
{
    float prev_x = x;

    if(div_x == 0 ) div_x=14;

    //   frame
    frame+=td*14;
    if( div_x > 0 ) {
        if( frame > 2) frame=0;
    }
    else {
        if( frame > 4) frame=2;
    }

    // 
    x+=div_x*td;

    if( is_IntoGround() || is_NoGround() )
    {
        x=prev_x;
        div_x *= (-1);
    }
}

void GO_Enemy1::CheckForHits()
{
    //      
    if( HitObject( player ) )
        GameControl->PlayerKilled();
}

void GO_Enemy1::Message(void *)
{
}


/*==========================================================
        ***   KEY    ****/

GO_Key::GO_Key()
{
    len=5; hgt=8;
    Frames.AddFrames( gameArt, 3,100,len,hgt,1,1 );
}

void GO_Key::CheckForHits()
{
    // ,      
    if( HitObject( player ) )
    {
        //  ,      
        Sprites->RemoveSprite(this);
        //   
        delete this;
        //   
        GameControl->keys++;
        //  ,   
        GameControl->RefreshGameBar();

        //    ,    
        if( GameControl->keys == GameControl->need_keys )
            GameControl->NextLevel();
        else
            //     ,    
            sampleMy->Play(0,0);
    }
}


/*==========================================================
        ***        ****/

GO_Kaplya_Emiter::GO_Kaplya_Emiter()
{
    len=6; hgt=2;
    Frames.AddFrames( gameArt, 18,100,len,hgt,1,1 );
    delay=delay=1+rand()%40/10.;
}
void GO_Kaplya_Emiter::Move(double td)
{
    //  3 .  
    delay-=td;
    if( delay < 0 )
    {
        delay=1+rand()%40/10.;
        GameControl->AddSprite( new GO_Kaplya(this));
    }
}

/*==========================================================
        ***         ****/

GO_Kaplya::GO_Kaplya(GO_Kaplya_Emiter *emiter)
{
    len=4; hgt=5;
    x=emiter->x;
    y=emiter->y;
    Frames.AddFrames( gameArt,31,101,len,hgt,1,1 );
    type=GO_KAPLYA;
}
void GO_Kaplya::Move(double td)
{
    div_y+=td*10;
    y+=div_y;
}

void GO_Kaplya::Message(void *m)
{
    Game_Message *msg=(Game_Message *)m;

    //   ,  
    if( msg->type == GM_INTOGROUND )
    {
        GameControl->RemoveSprite(this);
        delete this;
    }
}

void GO_Kaplya::CheckForHits()
{
    //   ,  
    if( HitObject( player ) )
    {
        GameControl->PlayerKilled();
    }

    //    ,   
    if( is_IntoGround() )
    {
        Game_Message msg={GM_INTOGROUND,0,this};
        Message(&msg);
    }
}

/*==========================================================
        ***        ****/

GO_Lift::GO_Lift()
{
    len=17; hgt=3;
    Frames.AddFrames( gameArt,40,102,len,hgt,1,1 );
    delay=0; action=-255;  // 1-  -1  -255-  
}
void GO_Lift::Search_Target()
{
    begin_x=x; begin_y=y; //   
    lin_pos = 0;    //     

    //  
    target_x=-255;
    Game_Object *o;
	for( int n=0; n<Sprites->NumSprites; n++)
	{
        o=(Game_Object*) Sprites->Sprites[n];
        if( o->type == GO_LIFT_TARGET )
        {
            // ,     (param    )
            if( o->param == param )
            {
                target_x = o->x;
                target_y = o->y;
            }
        }
    }
    //     - 
    if(target_x== (-255) )
    {
        OutMessage("Level Error: Lift "+IntToStr(param)+" without target!");
        GameControl->RemoveSprite(this);
        delete this;
    }
    else
    {
        // 
        track_len = sqrt( (target_y-begin_y)*(target_y-begin_y) +
                          (target_x-begin_x)*(target_x-begin_x) );
        //  
        cosk = (target_x-begin_x) / track_len;
        sink = (target_y-begin_y) / track_len;


        action=1;
    }
}

/*   
    action    direction, ..   
*/
void GO_Lift::Move(double td)
{
    //    action=-255
    //    ,     
    if( action == (-255) )
    {
        Search_Target();
        delay = 3;  //  
    }

    //  delay>0  
    if( delay>0 ) {
        delay-=td;  //   
        if( delay<0) delay=0;
        else return;
    }

    //   
    prev_lin_pos = lin_pos;

    //    
    lin_pos += action*td*50;

    //     
    //  ,   
    if( action < 0 && lin_pos<0 )
    {
        lin_pos = 0;
        action*= (-1);  //    ,   
        delay=1.5;  //      
    }

    if( action > 0 && lin_pos > track_len)
    {
        lin_pos = track_len;
        action*= (-1);  //    ,   
        delay = 1.5;  //      
    }

    //   
    prev_x = x;
    prev_y = y;

    //    
    //   ,    
    x = begin_x + lin_pos*cosk;
    y = begin_y + lin_pos*sink;

    //      ,
    if( player->lift == this )
    {
        //  ,   
        //  ,     
        div_x = x - prev_x;
        div_y = y - prev_y;

        //      
        Game_Message msg={GM_LIFTMOVE,0,this};
        player->Message(&msg);
    }

}

void GO_Lift::CheckForHits()
{
    //      
    // (,  4   )
    RECT pl = { player->x, player->y+player->hgt-4,
                player->x+player->len, player->y+player->hgt};
    RECT l = {x,y,x+len,y+hgt};

    if( !player->bLiftUse && HitRect(&pl,&l) )
    {
        //      
        player->y = y - player->hgt;
        player->action = SA_STILL;
        player->div_y = 0;
        player->frame = 0;
        player->bLiftUse = true;
        player->lift = this;
OutMessage("bLiftUse=true");
    }

    //     ,     ,
    //  ,     
    if( player->lift == this )
    {
        if( player->y != y - player->hgt ||
            player->x + player->len < x                ||
            player->x > x+len )
            {
                player->bLiftUse=false;
                player->lift = NULL;
OutMessage("Player is out of lift");
            }
    }
}

void GO_Lift::Message(void *m)
{
    Game_Message *msg=(Game_Message *)m;

    //   ,  
    if( msg->type == GM_CANTMOVE )
    {
        x = prev_x;
        y = prev_y;
        lin_pos = prev_lin_pos;
    }
}



/*==========================================================
        ***         ****/

GO_Lift_Target::GO_Lift_Target()
{
    type = GO_LIFT_TARGET;
    frame=-1; //  
}

